-------------------------------------------------------------
-- MSS copyright 2017
--	Filename:  AGC21.VHD
-- Author: Alain Zarembowitch / MSS
--	Version: 1
--	Date last modified: 11/16/17
-- Inheritance: 	AGC20.VHD 7/1/15
--
-- Description:  burst demodulator AGC 
-- Algorithm: track the input magnitude envelope with a fast upwards response and slow decay, 
-- then scale the input with 1/average magnitude.
-- Sharp gain adjustment when a burst comes in, slow decay until the next burst.
--
-- Rev 1 11/16/17 AZ
-- Added support for 16-bit ADCs
---------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

entity AGC21 is
	generic (
		CLK_FREQUENCY: integer := 160;
			-- CLK frequency in MHz. Needed to compute actual delays.
		SIMULATION: std_logic := '0'
	);
    port ( 

	--// GLOBAL CLOCKS
   CLK : in std_logic;				-- System clock
	SYNC_RESET: in std_logic;				-- synchronous reset

	--// INPUT DATA STREAM
	DATA_I_IN: in std_logic_vector(15 downto 0);
	DATA_Q_IN: in std_logic_vector(15 downto 0);
	SAMPLE_CLK_IN: in std_logic;
		-- sample clock. 

	--// CONTROLS 
	MTARGET: in std_logic_vector(13 downto 0);
		-- magnitude target for the output samples. Unsigned.
	CONTROL: in std_logic_vector(0 downto 0);
		-- bit 0: enabled(1) or bypassed(0)
	SYMBOLx2_CLK_IN: in std_logic;
		-- defines the AGC response time: approximately 2*modulation symbol rate is best

 	--// NORMALIZED OUTPUTS
 	-- latency: 3 CLK
	DATA_I_OUT: out std_logic_vector(13 downto 0);
	DATA_Q_OUT: out std_logic_vector(13 downto 0);
	SAMPLE_CLK_OUT: out std_logic;
		-- read at rising edge of CLK when SAMPLE_CLK_OUT = '1'
	SATURATION: out std_logic;
	
	--// TEST POINTS
	DATA_M : OUT std_logic_vector(13 downto 0);
	SAMPLE_M_CLK : OUT std_logic;
		-- input magnitude
	TP: out std_logic_vector(10 downto 1);
	CS1: out std_logic_vector(7 downto 0);
	CS1_CLK: out std_logic
		-- internal gain


			);
end entity;

architecture behavioral of AGC21 is
--------------------------------------------------------
--      COMPONENTS
--------------------------------------------------------
	COMPONENT POLAR3
	PORT(
		CLK : IN std_logic;
		X_IN : IN std_logic_vector(11 downto 0);
		Y_IN : IN std_logic_vector(11 downto 0);
		SAMPLE_CLK_IN : IN std_logic;          
		PHASE_OUT : OUT std_logic_vector(10 downto 0);
		MAGNITUDE_OUT : OUT std_logic_vector(13 downto 0);
		SAMPLE_CLK_OUT : OUT std_logic
		);
	END COMPONENT;

	COMPONENT TIMER_4US
	GENERIC (
		CLK_FREQUENCY: integer 
	);
	PORT(
		CLK : IN std_logic;          
		SYNC_RESET : IN std_logic;
		TICK_4US : OUT std_logic;
		TICK_100MS: out std_logic
		);
	END COMPONENT;

    COMPONENT DIVIDER
	 Generic(
		NBITS: integer
	 );
    PORT(
         CLK : IN  std_logic;
         NUM : IN  std_logic_vector((NBITS-1) downto 0);
         DENOM : IN  std_logic_vector((NBITS-1) downto 0);
         SAMPLE_CLK_IN : IN  std_logic;
         QUOTIENT : OUT  std_logic_vector((NBITS-1) downto 0);
         REMAINDER : OUT  std_logic_vector((NBITS-1) downto 0);
         SAMPLE_CLK_OUT : OUT  std_logic
        );
    END COMPONENT;

	component SIM2OUTFILE is
		 Generic (
			FILENAME: string
		 );
		 Port ( 
			--// CLK, RESET
			CLK: in std_logic;
				-- Reference clock
			DATA1: in std_logic_vector(11 downto 0);
			DATA2: in std_logic_vector(11 downto 0);
			DATA3: in std_logic_vector(11 downto 0);
			SAMPLE_CLK_IN: in std_logic
				-- DATA_IN is read at the rising edge of CLK when SAMPLE_CLK_IN = '1'

	);
	END COMPONENT;
--------------------------------------------------------
--     SIGNALS
--------------------------------------------------------
-- Suffix _D indicates a one CLK delayed version of the net with the same name
-- Suffix _E indicates an extended precision version of the net with the same name
-- Suffix _N indicates an inverted version of the net with the same name

signal TICK_100MS: std_logic := '0';
signal DATA1M: std_logic_vector(13 downto 0) := (others => '0');
signal SAMPLE1M_CLK: std_logic := '0';
signal DATA1M_ENVELOPE: unsigned(13 downto 0) := (others => '0');
signal DATA1M_ENVELOPEx: unsigned(19 downto 0) := (others => '0');

signal ONEOVERX_DENOM: unsigned(23 downto 0) := (others => '0');
signal ONEOVERX_NUM: unsigned(23 downto 0) := (others => '0');
signal ONEOVERMAG1: std_logic_vector(23 downto 0) := (others => '0');
signal ONEOVERMAG1_SAMPLE_CLK: std_logic := '0';
signal DATA0_I: signed(17 downto 0) := (others => '0');
signal DATA0_Q: signed(17 downto 0) := (others => '0');
signal SAMPLE0_CLK: std_logic := '0';
signal DATA2_I: signed(35 downto 0) := (others => '0');
signal DATA2_Q: signed(35 downto 0) := (others => '0');
signal DATA2_I_D: signed(35 downto 0) := (others => '0');
signal DATA2_Q_D: signed(35 downto 0) := (others => '0');
signal SAMPLE2_CLK: std_logic := '0';
signal INTERNAL_GAIN: signed(17 downto 0) := (others => '0');
signal INTERNAL_GAIN_SHIFT: std_logic := '0';
signal INTERNAL_GAIN_SHIFT_D: std_logic := '0';
signal DATA3_I: signed(13 downto 0) := (others => '0');
signal DATA3_Q: signed(13 downto 0) := (others => '0');
signal SAMPLE3_CLK: std_logic := '0';
signal SATURATION_AGC_I: std_logic := '0';
signal SATURATION_AGC_Q: std_logic := '0';
signal SATURATION_local: std_logic := '0';
--------------------------------------------------------
--      IMPLEMENTATION
--------------------------------------------------------

begin

-- measure of time. One pulse every 100ms. Used to compute probability of false alarm in this period.
TIMER_4US_NOSIM: if(SIMULATION = '0') generate
	TIMER_4US_001: TIMER_4US 
	GENERIC MAP(
		CLK_FREQUENCY => CLK_FREQUENCY	--  in MHz
	)
	PORT MAP(
		CLK => CLK,
		SYNC_RESET => SYNC_RESET,
		TICK_4US => open,
		TICK_100MS => TICK_100MS
	);
end generate;
TIMER_4US_SIM: if(SIMULATION = '1') generate
	TIMER_4US_001: process
	begin
		wait for 3.125ns;
		TICK_100MS <= '0';
		wait for 200us;
		TICK_100MS <= '1';
		wait for 3.125ns;
	end process;
end generate;

--//-- INTERNAL AGC --------------------------------------------------------------
-- Feedforward control: input is scaled by 1/average magnitude

-- Complex/Real input to magnitude
-- Latency: 5 CLK
POLAR3_001: POLAR3 PORT MAP(
	CLK => CLK,
	X_IN => DATA_I_IN(15 downto 4),
	Y_IN => DATA_Q_IN(15 downto 4),
	SAMPLE_CLK_IN => SAMPLE_CLK_IN,
	PHASE_OUT => open,	-- for simulation, unused (optimized out) during build.
	MAGNITUDE_OUT => DATA1M,	-- sqrt(X_IN(13:0)^2 + Y_IN(13:0)^2)
	SAMPLE_CLK_OUT => SAMPLE1M_CLK 
);
-- monitoring
DATA_M <= DATA1M;
SAMPLE_M_CLK <= SAMPLE1M_CLK;

-- envelope
ENVELOPE_001: process(CLK)
begin
	if rising_edge(CLK) then
		if(SYNC_RESET = '1') then
			DATA1M_ENVELOPEx <= (others => '0');
		elsif(SAMPLE1M_CLK = '1') and (unsigned(DATA1M) > DATA1M_ENVELOPE) then
			-- fast rise
			DATA1M_ENVELOPEx <= DATA1M_ENVELOPEx + resize(unsigned(DATA1M),20);	-- add 1/64
		elsif(TICK_100MS = '1') then
			-- slow decay (6.4 seconds halflife)
			DATA1M_ENVELOPEx <= DATA1M_ENVELOPEx  - resize(DATA1M_ENVELOPE,20);	-- remove 1/64
		end if;
	end if;
end process;
DATA1M_ENVELOPE <= unsigned(DATA1M_ENVELOPEx(19 downto 6));

-- inverse magnitude
ONEOVERX_NUM <= unsigned(MTARGET & "0000000000");	-- extend target magnitude with 10 LSbs
ONEOVERX_DENOM <= resize(DATA1M_ENVELOPE, 24); -- extend average to 24 bits
ONEOVERX_001: DIVIDER 
GENERIC MAP(
	NBITS => 24
)
PORT MAP (
		 CLK => CLK,
		 NUM => std_logic_vector(ONEOVERX_NUM),
		 DENOM => std_logic_vector(ONEOVERX_DENOM),
		 SAMPLE_CLK_IN => SYMBOLx2_CLK_IN,	
		 QUOTIENT => ONEOVERMAG1,	-- format 14.10
		 REMAINDER => open,
		 SAMPLE_CLK_OUT => ONEOVERMAG1_SAMPLE_CLK
	  );


-- reclock at multiplier input for best timing
INTERNAL_AGC_001: process(CLK)
begin
	if rising_edge(CLK) then
		DATA0_I <= signed(DATA_I_IN & "00");	-- extend precision for 18x18 multiplier
		DATA0_Q <= signed(DATA_Q_IN & "00");
		SAMPLE0_CLK <= SAMPLE_CLK_IN;
		
		if(CONTROL(0) = '1') then
			-- internal gain control is enabled.
			if(unsigned(ONEOVERMAG1(23 downto 17)) /= 0) then
				INTERNAL_GAIN_SHIFT <= '1';	-- shift result by 7 bits
				INTERNAL_GAIN <= signed("0" & ONEOVERMAG1(23 downto 7));				
			else
				INTERNAL_GAIN_SHIFT <= '0';
				INTERNAL_GAIN <= signed("0" & ONEOVERMAG1(16 downto 0));
				-- gain. Always positive.
			end if;
		else
			-- bypass. Fixed unit gain here.
			INTERNAL_GAIN <= "000000010000000000";	-- verified by sim 12/23/12 az
		end if;
	end if;
end process;

DATA2_I <=  DATA0_I * INTERNAL_GAIN;	-- format 1.17 * 8.10
DATA2_Q <=  DATA0_Q * INTERNAL_GAIN;	-- format 1.17 * 8.10

-- reclock after multiplier for better timing
INTERNAL_AGC_004: process(CLK)
begin
	if rising_edge(CLK) then
		SAMPLE2_CLK <= SAMPLE0_CLK;
		INTERNAL_GAIN_SHIFT_D <= INTERNAL_GAIN_SHIFT;	
		DATA2_I_D <= DATA2_I;
		DATA2_Q_D <= DATA2_Q;
	end if;
end process;

-- AGC range > 42 dB (7 bits) (TBC)
INTERNAL_AGC_005: process(CLK)
begin
	if rising_edge(CLK) then
		SAMPLE3_CLK <= SAMPLE2_CLK;
		SATURATION_local <= SATURATION_AGC_I or SATURATION_AGC_Q;
		
		if(INTERNAL_GAIN_SHIFT_D = '0') then
			if(DATA2_I_D(35) = '1') and (DATA2_I_D(34 downto 27) /= "11111111") then
				-- negative saturation
				DATA3_I <= "10000000000001";
				SATURATION_AGC_I <= '1';
			elsif(DATA2_I_D(35) = '0') and (DATA2_I_D(34 downto 27) /= 0) then
				-- positive saturation
				DATA3_I <= "01111111111111";
				SATURATION_AGC_I <= '1';
			else
				DATA3_I <= DATA2_I_D(27 downto 14);	
				SATURATION_AGC_I <= '0';
			end if;
			if(DATA2_Q_D(35) = '1') and (DATA2_Q_D(34 downto 27) /= "11111111") then
				-- negative saturation
				DATA3_Q <= "10000000000001";
				SATURATION_AGC_Q <= '1';
			elsif(DATA2_Q_D(35) = '0') and (DATA2_Q_D(34 downto 27) /= 0) then
				-- positive saturation
				DATA3_Q <= "01111111111111";
				SATURATION_AGC_Q <= '1';
			else
				DATA3_Q <= DATA2_Q_D(27 downto 14);	
				SATURATION_AGC_Q <= '0';
			end if;
		else
			-- restore the 7-bit shift
			if(DATA2_I_D(35) = '1') and ((not DATA2_I_D(34 downto 20)) /= 0) then
				-- negative saturation
				DATA3_I <= "10000000000001";
				SATURATION_AGC_I <= '1';
			elsif(DATA2_I_D(35) = '0') and (DATA2_I_D(34 downto 20) /= 0) then
				-- positive saturation
				DATA3_I <= "01111111111111";
				SATURATION_AGC_I <= '1';
			else
				DATA3_I <= DATA2_I_D(20 downto 7);	
				SATURATION_AGC_I <= '0';
			end if;
			if(DATA2_Q_D(35) = '1') and ((not DATA2_Q_D(34 downto 20)) /= 0) then
				-- negative saturation
				DATA3_Q <= "10000000000001";
				SATURATION_AGC_Q <= '1';
			elsif(DATA2_Q_D(35) = '0') and (DATA2_Q_D(34 downto 20) /= 0) then
				-- positive saturation
				DATA3_Q <= "01111111111111";
				SATURATION_AGC_Q <= '1';
			else
				DATA3_Q <= DATA2_Q_D(20 downto 7);
				SATURATION_AGC_Q <= '0';
			end if;
		end if;
	end if;
end process;
SATURATION <= SATURATION_local;

--// output
DATA_I_OUT <= std_logic_vector(DATA3_I);
DATA_Q_OUT <= std_logic_vector(DATA3_Q);
SAMPLE_CLK_OUT <= SAMPLE3_CLK;

--// ComScope monitoring
CS1 <= std_logic_vector(INTERNAL_GAIN(13 downto 6));
CS1_CLK <= ONEOVERMAG1_SAMPLE_CLK;

--// simulation. send to a tab-delimited text file for subsequent (Matlab) plot.
--	SIM2OUTFILE_002: SIM2OUTFILE 
--	generic map(
--		FILENAME => "AGC20b.txt"
--	)
--	PORT MAP(
--		CLK => CLK,
--		DATA1 => INTERNAL_GAIN(17 downto 6),
--		DATA2 => INTERNAL_GAIN(14 downto 3),
--		DATA3 => INTERNAL_GAIN(11 downto 0),
--		SAMPLE_CLK_IN => '1'
--	);
--end generate;

end behavioral;
