-------------------------------------------------------------
-- Filename:  CPM_PHASEx4.VHD
-- Authors: 
--		Alain Zarembowitch / MSS
-- Version: Rev 1
-- Last modified: 10/25/17

-- Inheritance: 	N/A
--
-- description:  synthesizable ROM containing the cumulative phase over ONE symbol T 
-- for various CPM modulations and last L symbol sequences. 
-- All are sampled at 4 samples/symbol, 12-bit unsigned representing the [0-2pi[ range.
-- assuming a modulation index h = 0.5.
-- The table is constructed for the most complex case of LMAX = 4. 
-- The address is constructed as
	-- bits (1:0)  sample number, range [0 - 3[
	-- bits (5:2)  symbols assumption for most complex case M^L = 16, range [0 - 16[
	-- 				LSb is the oldest bit
	-- bits (9:6) modulation scheme
			-- All assume h = 0.5 and M = 2 unless otherwise stated
			-- 0 = FSK  
			-- 1 = PCM/FM
			-- 2 = GFSK BT=0.7	
			-- 3 = GFSK BT=0.5	
			-- 4 = GFSK BT=0.3	
			-- 5 = SOQPSK-MIL (M=3)
			-- 6 = SOQPSK-TG	(M=3)
			-- 7 = multi-h ARTM CPM (Tier 2) 
			-- 8 = GFSK BT=0.25
-- Table entries were generated using the demod_fsk1.m Matlab program (in the /Matlab folder)
---------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use ieee.numeric_std.all;

entity CPM_PHASEx4 is
	 Generic (
		DATA_WIDTH: integer := 12;	
		ADDR_WIDTH: integer := 10	
	);
    Port ( 
	    -- Port A
		CLKA   : in  std_logic;
		CSA: in std_logic;	-- chip select, active high
		OEA : in std_logic;	-- output enable, active high
		ADDRA  : in  std_logic_vector(ADDR_WIDTH-1 downto 0);
		DOA  : out std_logic_vector(DATA_WIDTH-1 downto 0);

		-- Port B
		CLKB   : in  std_logic;
		CSB: in std_logic;	-- chip select, active high
		OEB : in std_logic;	-- output enable, active high
		ADDRB  : in  std_logic_vector(ADDR_WIDTH-1 downto 0);
		DOB  : out std_logic_vector(DATA_WIDTH-1 downto 0)
		);
end entity;

architecture Behavioral of CPM_PHASEx4 is
--------------------------------------------------------
--     SIGNALS
--------------------------------------------------------
-- inferred rom
signal DOA_local: std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
signal DOB_local: std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
type ROM_TYPE is array (0 to (2**ADDR_WIDTH)-1) of std_logic_vector(DATA_WIDTH-1 downto 0);
constant ROM : ROM_TYPE := (
-- FSK (L = 1)
x"F00", x"E00", x"D00", x"C00", x"100", x"200", x"300", x"400",	-- xxx0, xxx1
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0110
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1110
-- PCM/FM (L = 2)
-- sequence hypothesis order: LSb is the oldest received bit (same as in demod_pcmfm_baseline.m
x"f00", x"e00", x"d00", x"c00", x"0ED", x"14E", x"0ED", x"000", 	-- 00,01
x"f13", x"EB2", x"F13", x"000", x"100", x"200", x"300", x"400",	-- 10,11
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
-- GFSK BT=0.7 truncated frequency pulse over T (1 bit period)
x"FEF", x"F0E", x"CF2", x"C11", x"011", x"0F2", x"30E", x"3EF",	-- -1,+1
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
-- GFSK BT=0.5 truncated frequency pulse over T (1 bit period)
x"FB8", x"EC1", x"D3F", x"C48", x"048", x"13F", x"2C1", x"3B8",  	-- -1,+1
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
-- GFSK BT=0.3 truncated frequency pulse over T (1 bit period)
x"F39", x"E74", x"D8D", x"CC8", x"0C7", x"18C", x"273", x"338",  	-- -1,+1
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
-- SOQPSK-MIL  
x"F00", x"E00", x"D00", x"C00", x"000", x"000", x"000", x"000",	-- -1, 0
x"100", x"200", x"300", x"400", x"000", x"000", x"000", x"000",	-- +1	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
-- SOQPSK-TG truncated frequency pulse over T (1 bit period)
x"EE7", x"E50", x"DB0", x"D19", x"000", x"000", x"000", x"000",	-- -1, 0
x"119", x"1B0", x"250", x"2E7", x"000", x"000", x"000", x"000",	-- +1	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
-- multi-h CPM (ARTM Tier 2) 
x"C00", x"800", x"400", x"000", x"C00", x"800", x"400", x"000",	-- 0000
x"400", x"800", x"400", x"000", x"C00", x"800", x"400", x"000",	-- 1000
x"C00", x"800", x"C00", x"000", x"C00", x"800", x"400", x"000",	-- 0100
x"C00", x"800", x"400", x"000", x"400", x"800", x"400", x"000",	-- 1100
x"C00", x"800", x"400", x"000", x"400", x"800", x"400", x"000",	-- 0010
x"C00", x"800", x"400", x"000", x"400", x"800", x"400", x"000",	-- 1010
x"C00", x"800", x"400", x"000", x"400", x"800", x"400", x"000",	-- 0110
x"C00", x"800", x"400", x"000", x"400", x"800", x"400", x"000",	-- 1110
-- GFSK BT=0.25 truncated frequency pulse over T (1 bit period)
x"F11", x"E65", x"DA4", x"CF8", x"0EF", x"19B", x"25C", x"308",  	-- -1,+1
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
-- future
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0110
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1110
-- future
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0110
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1110
-- future
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0110
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1110
-- future
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0110
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1110
-- future
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0110
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1110
-- future
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0110
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1110
-- future
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0110
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000"	-- 1110
);


--------------------------------------------------------
--      IMPLEMENTATION
--------------------------------------------------------
begin

-- Port A read
process(CLKA)
begin
	if rising_edge(CLKA) then
		if(CSA = '1') then
			DOA_local <= ROM(to_integer(unsigned(ADDRA)));
		else
			DOA_local <= (others => '0');
		end if;
	end if;
end process;

-- tri-state output
DOA <= DOA_local when (CSA = '1') and (OEA = '1') else (others => 'Z');
 

-- Port B read
process(CLKB)
begin
	if rising_edge(CLKB) then
		if(CSB = '1') then
			DOB_local <= ROM(to_integer(unsigned(ADDRB)));
		else
			DOB_local <= (others => '0');
		end if;
	end if;
end process;

-- tri-state output
DOB <= DOB_local when (CSB = '1') and (OEB = '1') else (others => 'Z');

end Behavioral;
