-------------------------------------------------------------
-- Filename:  CPM_PHASEx4.VHD
-- Authors: 
--		Alain Zarembowitch / MSS
-- Version: Rev 1
-- Last modified: 9/13/17

-- Inheritance: 	N/A
--
-- description:  synthesizable ROM containing the cumulative phase over ONE symbol T 
-- for various CPM modulations and last L symbol sequences. 
-- All are sampled at 4 samples/symbol, 12-bit unsigned representing the [0-2pi[ range.
-- assuming a modulation index h = 0.5.
-- The table is constructed for the most complex case of LMAX = 4. 
-- The address is constructed as
	-- case FSK, M=2, L=1 start address offset 0
		-- bits (1:0)  sample number, range [0 - 3[  (4 samples/symbols * L)
		-- bits (2:2)  symbols assumptions M^L = 2 
		-- bits (9:3)  0
	-- case PCM/FM, M=2, L=2 start address offset 8
		-- bits (2:0)  sample number, range [0 - 7[  (4 samples/symbols * L)
		-- bits (4:3)  symbols assumptions M^L = 4
		-- bits (9:5)  0
		
-- Table entries were generated using the demod_fsk1.m Matlab program (in the /Matlab folder)
---------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use ieee.numeric_std.all;

entity CPM_PHASEx4 is
	 Generic (
		DATA_WIDTH: integer := 12;	
		ADDR_WIDTH: integer := 10	
	);
    Port ( 
	    -- Port A
		CLKA   : in  std_logic;
		CSA: in std_logic;	-- chip select, active high
		OEA : in std_logic;	-- output enable, active high
		ADDRA  : in  std_logic_vector(ADDR_WIDTH-1 downto 0);
		DOA  : out std_logic_vector(DATA_WIDTH-1 downto 0);

		-- Port B
		CLKB   : in  std_logic;
		CSB: in std_logic;	-- chip select, active high
		OEB : in std_logic;	-- output enable, active high
		ADDRB  : in  std_logic_vector(ADDR_WIDTH-1 downto 0);
		DOB  : out std_logic_vector(DATA_WIDTH-1 downto 0)
		);
end entity;

architecture Behavioral of CPM_PHASEx4 is
--------------------------------------------------------
--     SIGNALS
--------------------------------------------------------
-- inferred rom
signal DOA_local: std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
signal DOB_local: std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
type ROM_TYPE is array (0 to (2**ADDR_WIDTH)-1) of std_logic_vector(DATA_WIDTH-1 downto 0);
constant ROM : ROM_TYPE := (
-- FSK (L = 1)
x"F00", x"E00", x"D00", x"C00", x"100", x"200", x"300", x"400",	-- xxx0, xxx1
-- PCM/FM (L = 2)
x"FF6", x"FA7", x"EF6", x"E00", x"D0A", x"C59", x"C0A", x"C00",	-- 00
x"00a", x"059", x"10a", x"200", x"10a", x"059", x"00a", x"000",	-- 01
x"FF6", x"fa7", x"EF6", x"e00", x"EF6", x"fa7", x"FF6", x"000",	-- 10
x"00a", x"059", x"10a", x"200", x"2f6", x"3a7", x"3f6", x"400",	-- 11
-- GFSK BT=0.7
x"DE4", x"d03", x"CE2", x"c00", x"de4", x"D03", x"D03", x"de4",	-- 00,
x"21C", x"2fd", x"2df", x"21c", x"21C", x"2fd", x"31e", x"400",	-- .., 11
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
-- GFSK BT=0.5
x"f7e", x"e80", x"cfe", x"c00", x"f7e", x"e80", x"cff", x"c0f",	-- 000, ...
x"000", x"0f0", x"271", x"361", x"000", x"0f0", x"271", x"36f",	-- 
x"000", x"f10", x"d8f", x"c91", x"000", x"f10", x"d8f", x"c9f",	--
x"082", x"180", x"301", x"3f1", x"082", x"180", x"302", x"400",	-- ..., 111
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
-- GFSK BT=0.3
x"F0C", x"e0c", x"D01", x"c01", x"f0d", x"E16", x"d2e", x"c9c",	-- 0000
x"fff", x"089", x"14d", x"1d7", x"001", x"093", x"17a", x"272",	-- 1000
x"FFF", x"f6d", x"e86", x"D8E", x"001", x"f77", x"EB3", x"e29",	-- 0100
x"0f3", x"1EA", x"2d2", x"364", x"0F4", x"1f4", x"2ff", x"3ff",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
-- SOQPSK-MIL  
x"F00", x"E00", x"D00", x"C00", x"000", x"000", x"000", x"000",	-- -1,0,
x"100", x"200", x"300", x"400", x"000", x"000", x"000", x"000",	-- 1
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
-- SOQPSK-TG (TODO !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!)
x"C00", x"800", x"400", x"000", x"C00", x"800", x"400", x"000",	-- 0000
x"400", x"800", x"400", x"000", x"C00", x"800", x"400", x"000",	-- 1000
x"C00", x"800", x"C00", x"000", x"C00", x"800", x"400", x"000",	-- 0100
x"C00", x"800", x"400", x"000", x"400", x"800", x"400", x"000",	-- 1100
x"C00", x"800", x"400", x"000", x"400", x"800", x"400", x"000",	-- 0010
x"C00", x"800", x"400", x"000", x"400", x"800", x"400", x"000",	-- 1010
x"C00", x"800", x"400", x"000", x"400", x"800", x"400", x"000",	-- 0110
x"C00", x"800", x"400", x"000", x"400", x"800", x"400", x"000",	-- 1110
-- multi-h CPM (ARTM Tier 2) 
x"C00", x"800", x"400", x"000", x"C00", x"800", x"400", x"000",	-- 0000
x"400", x"800", x"400", x"000", x"C00", x"800", x"400", x"000",	-- 1000
x"C00", x"800", x"C00", x"000", x"C00", x"800", x"400", x"000",	-- 0100
x"C00", x"800", x"400", x"000", x"400", x"800", x"400", x"000",	-- 1100
x"C00", x"800", x"400", x"000", x"400", x"800", x"400", x"000",	-- 0010
x"C00", x"800", x"400", x"000", x"400", x"800", x"400", x"000",	-- 1010
x"C00", x"800", x"400", x"000", x"400", x"800", x"400", x"000",	-- 0110
x"C00", x"800", x"400", x"000", x"400", x"800", x"400", x"000",	-- 1110
-- GFSK BT=0.25
x"F06", x"e06", x"D05", x"c06", x"f0d", x"E25", x"d65", x"cf3",	-- 0000
x"FFC", x"056", x"0d6", x"12F", x"004", x"075", x"136", x"21d",	-- 1000
x"ffc", x"F8B", x"eca", x"de3", x"004", x"faa", x"f2a", x"ED1",	-- 0100
x"0f3", x"1db", x"29B", x"30d", x"0fa", x"1FA", x"2fb", x"3fa",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
-- future
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0110
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1110
-- future
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0110
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1110
-- future
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0110
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1110
-- future
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0110
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1110
-- future
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0110
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1110
-- future
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0110
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1110
-- future
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0110
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1110
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1000
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1100
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 1010
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000",	-- 0110
x"000", x"000", x"000", x"000", x"000", x"000", x"000", x"000"	-- 1110
);


--------------------------------------------------------
--      IMPLEMENTATION
--------------------------------------------------------
begin

-- Port A read
process(CLKA)
begin
	if rising_edge(CLKA) then
		if(CSA = '1') then
			DOA_local <= ROM(to_integer(unsigned(ADDRA)));
		else
			DOA_local <= (others => '0');
		end if;
	end if;
end process;

-- tri-state output
DOA <= DOA_local when (CSA = '1') and (OEA = '1') else (others => 'Z');
 

-- Port B read
process(CLKB)
begin
	if rising_edge(CLKB) then
		if(CSB = '1') then
			DOB_local <= ROM(to_integer(unsigned(ADDRB)));
		else
			DOB_local <= (others => '0');
		end if;
	end if;
end process;

-- tri-state output
DOB <= DOB_local when (CSB = '1') and (OEB = '1') else (others => 'Z');

end Behavioral;
