-------------------------------------------------------------
-- Filename:  CPM_FILTERSx4.VHD
-- Authors: 
-- 	from http://www.xilinx.com/support/documentation/sw_manuals/xilinx14_4/xst_v6s6.pdf  p262
--		Alain Zarembowitch / MSS
-- Version: Rev 1
-- Last modified: 8/9/17

-- Inheritance: 	N/A
--
-- description:  synthesizable ROM containing the frequency impulses (premod filters) for various CPM modulations
-- All are sampled at 4 samples/symbol, 16-bit signed (2's complement), fixed-point format 1.15
-- For 8x sampling, see CPM_FILTERSx8.VHD
-- Use MATLAB siggen_fsk1.m to generate the ROM contents.
--
-- Modulation   Start address  Number of samples filter length(L)
-- FSK	x0	4 1 rectangular frequency impulse
-- PCM/FM x4	7 2 raised cosine premod filter
-- GFSK BT=0.7 x0C  7 2
-- GFSK BT=0.5 x14  9 3(rounded up to 4 when implementing the modulation)
-- GFSK BT=0.3 x24  15 4 
-- SOQPSK-MIL x34  4 2
-- SOQPSK-TG x38  31 8
-- multi-h CPM (ARTM Tier 2) x58 11 4
-- GFSK BT=0.25 x68  19 5 (rounded up to 8 when implementing the modulation)
--
-- Rev 1 8/9/17 AZ
-- Added GFSK BT=0.25
-- Corrected SOQPSK-TG (normalized)
---------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use ieee.numeric_std.all;

entity CPM_FILTERSx4 is
	 Generic (
		DATA_WIDTH: integer := 16;	-- FIXED!
		ADDR_WIDTH: integer := 8	-- defines ROM maximum size
	);
    Port ( 
	    -- Port A
		CLKA   : in  std_logic;
		CSA: in std_logic;	-- chip select, active high
		OEA : in std_logic;	-- output enable, active high
		ADDRA  : in  std_logic_vector(ADDR_WIDTH-1 downto 0);
		DOA  : out std_logic_vector(DATA_WIDTH-1 downto 0);

		-- Port B
		CLKB   : in  std_logic;
		CSB: in std_logic;	-- chip select, active high
		OEB : in std_logic;	-- output enable, active high
		ADDRB  : in  std_logic_vector(ADDR_WIDTH-1 downto 0);
		DOB  : out std_logic_vector(DATA_WIDTH-1 downto 0)
		);
end entity;

architecture Behavioral of CPM_FILTERSx4 is
--------------------------------------------------------
--     SIGNALS
--------------------------------------------------------
-- inferred rom
signal DOA_local: std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
signal DOB_local: std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
type ROM_TYPE is array (0 to (2**ADDR_WIDTH)-1) of std_logic_vector(DATA_WIDTH-1 downto 0);
constant ROM : ROM_TYPE := (
-- FSK
x"2000", x"2000", x"2000", x"2000", 
-- PCM/FM
x"04B0", x"1000", x"1B50", x"2000", 
x"1B50", x"1000", x"04B0", x"0000",
-- GFSK BT=0.7
x"0007", x"020F", x"1C32", x"4371", 
x"1C32", x"020F", x"0007", x"0000", 
-- GFSK BT=0.5
x"0000", x"0000", x"0000", x"000A", 
x"00E1", x"0820", x"1EDF", x"302C", 
x"1EDF", x"0820", x"00E1", x"000A", 
x"0000", x"0000", x"0000", x"0000", 
-- GFSK BT=0.3
x"0003", x"0017", x"0087", x"023A", 
x"06D6", x"0F3B", x"18A0", x"1CE7", 
x"18A0", x"0F3B", x"06D6", x"023A", 
x"0087", x"0017", x"0003", x"0000", 
-- SOQPSK-MIL
x"2000", x"2000", x"2000", x"2000", 
-- SOQPSK-TG 
-- corrected (sum = 1) *080917
x"fffc", x"fffe", x"0002", x"ffe1",
x"ff86", x"fefe", x"fe80", x"fe65",
x"ff13", x"00dc", x"03d9", x"07cd",
x"0c27", x"1020", x"12e9", x"13ea",
x"12e9", x"1020", x"0c27", x"07cd",
x"03d9", x"00dc", x"ff13", x"fe65",
x"fe80", x"fefe", x"ff86", x"ffe1",
x"0002", x"fffe", x"fffc", x"0000",
-- multi-h CPM (ARTM Tier 2) 
x"0000", x"0000", x"016E", x"0555",
x"0AAB", x"1000", x"13E7", x"1555",
x"13E7", x"1000", x"0AAB", x"0555",
x"016E", x"0000", x"0000", x"0000",
-- GFSK BT=0.25
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0001", x"0005",
x"001a", x"0070", x"017e", x"0410",
x"08da", x"0f70", x"158d", x"1816",
x"158d", x"0f70", x"08da", x"0410",
x"017e", x"0070", x"001a", x"0005",
x"0001", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
-- future
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000",
x"0000", x"0000", x"0000", x"0000"
);









--------------------------------------------------------
--      IMPLEMENTATION
--------------------------------------------------------
begin

-- Port A read
process(CLKA)
begin
	if rising_edge(CLKA) then
		if(CSA = '1') then
			DOA_local <= ROM(to_integer(unsigned(ADDRA)));
		else
			DOA_local <= (others => '0');
		end if;
	end if;
end process;

-- tri-state output
DOA <= DOA_local when (CSA = '1') and (OEA = '1') else (others => 'Z');
 

-- Port B read
process(CLKB)
begin
	if rising_edge(CLKB) then
		if(CSB = '1') then
			DOB_local <= ROM(to_integer(unsigned(ADDRB)));
		else
			DOB_local <= (others => '0');
		end if;
	end if;
end process;

-- tri-state output
DOB <= DOB_local when (CSB = '1') and (OEB = '1') else (others => 'Z');

end Behavioral;
