-------------------------------------------------------------
-- Filename:  CPM_FILTERSx8.VHD
-- Authors: 
-- 	from http://www.xilinx.com/support/documentation/sw_manuals/xilinx14_4/xst_v6s6.pdf  p262
--		Alain Zarembowitch / MSS
-- Version: Rev 0
-- Last modified: 8/28/16

-- Inheritance: 	N/A
--
-- description:  synthesizable ROM containing the frequency impulses (premod filters) for various CPM modulations
-- All are sampled at 8 samples/symbol, 16-bit signed (2's complement), fixed-point format 1.15
-- For 4x sampling, see CPM_FILTERSx4.VHD
-- Modulation   Start address  Number of samples filter length(L)
-- FSK	x0	8 1 rectangular frequency impulse
-- PCM/FM x8	17 2 raised cosine premod filter
-- GFSK BT=0.7 x18  13 2
-- GFSK BT=0.5 x28  19 3(rounded up to 4 when implementing the modulation)
-- GFSK BT=0.3 x48  29 4 
-- SOQPSK-MIL x68  15 2
-- SOQPSK-TG x70  63 8
-- multi-h CPM (ARTM Tier 2) xB0 3(rounded up to 4 when implementing the modulation)
---------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use ieee.numeric_std.all;

entity CPM_FILTERSx8 is
	 Generic (
		DATA_WIDTH: integer := 16;	-- FIXED!
		ADDR_WIDTH: integer := 8	-- defines ROM maximum size
	);
    Port ( 
	    -- Port A
		CLKA   : in  std_logic;
		CSA: in std_logic;	-- chip select, active high
		OEA : in std_logic;	-- output enable, active high
		ADDRA  : in  std_logic_vector(ADDR_WIDTH-1 downto 0);
		DOA  : out std_logic_vector(DATA_WIDTH-1 downto 0);

		-- Port B
		CLKB   : in  std_logic;
		CSB: in std_logic;	-- chip select, active high
		OEB : in std_logic;	-- output enable, active high
		ADDRB  : in  std_logic_vector(ADDR_WIDTH-1 downto 0);
		DOB  : out std_logic_vector(DATA_WIDTH-1 downto 0)
		);
end entity;

architecture Behavioral of CPM_FILTERSx8 is
--------------------------------------------------------
--     SIGNALS
--------------------------------------------------------
-- inferred rom
signal DOA_local: std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
signal DOB_local: std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
type ROM_TYPE is array (0 to (2**ADDR_WIDTH)-1) of std_logic_vector(DATA_WIDTH-1 downto 0);
constant ROM : ROM_TYPE := (
-- FSK
x"1000", x"1000", x"1000", x"1000", x"1000", x"1000", x"1000", x"1000",
-- PCM/FM
x"0000", x"009C", x"0258", x"04F0", x"0800", x"0B10", x"0DA8", x"0F64",
x"1000", x"0F64", x"0DA8", x"0B10", x"0800", x"04F0", x"0258", x"009C",
-- GFSK BT=0.7
x"0003", x"0025", x"0108", x"04BD", x"0E19", x"1B1D", x"21B8", x"1B1D", 
x"0E19", x"04BD", x"0108", x"0025", x"0003", x"0000", x"0000", x"0000", 
-- GFSK BT=0.5
x"0001", x"0005", x"001A", x"0070", x"017E", x"0410", x"08DA", x"0F70", 
x"158D", x"1816", x"158D", x"0F70", x"08DA", x"0410", x"017E", x"0070", 
x"001A", x"0005", x"0001", x"0000", x"0000", x"0000", x"0000", x"0000", 
x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", 
-- GFSK BT=0.3
x"0001", x"0004", x"000C", x"001D", x"0043", x"0090", x"011D", x"0208", 
x"036B", x"054F", x"079D", x"0A14", x"0C50", x"0DE2", x"0E74", x"0DE2", 
x"0C50", x"0A14", x"079D", x"054F", x"036B", x"0208", x"011D", x"0090", 
x"0043", x"001D", x"000C", x"0004", x"0001", x"0000", x"0000", x"0000", 
-- SOQPSK-MIL
x"1000", x"1000", x"1000", x"1000", x"1000", x"1000", x"1000", x"1000",
-- SOQPSK-TG 
x"FFFF", x"FFFE", x"FFFE", x"FFFF", x"0001", x"0001", x"FFFD", x"FFF3", 
x"FFE4", x"FFCF", x"FFB5", x"FF99", x"FF7D", x"FF66", x"FF59", x"FF5B", 
x"FF71", x"FFA1", x"FFED", x"0058", x"00E3", x"018C", x"024D", x"0322", 
x"0401", x"04E2", x"05B8", x"067A", x"071D", x"0799", x"07E6", x"0800", 
x"07E6", x"0799", x"071D", x"067A", x"05B8", x"04E2", x"0401", x"0322", 
x"024D", x"018C", x"00E3", x"0058", x"FFED", x"FFA1", x"FF71", x"FF5B", 
x"FF59", x"FF66", x"FF7D", x"FF99", x"FFB5", x"FFCF", x"FFE4", x"FFF3", 
x"FFFD", x"0001", x"0001", x"FFFF", x"FFFE", x"FFFE", x"FFFF", x"0000",
-- multi-h CPM (ARTM Tier 2) 
x"002F", x"00B7", x"0190", x"02AB", x"03F4", x"0555", x"06B7", x"0800", 
x"091B", x"09F4", x"0A7C", x"0AAB", x"0A7C", x"09F4", x"091B", x"0800", 
x"06B7", x"0555", x"03F4", x"02AB", x"0190", x"00B7", x"002F", x"0000", 
x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", x"0000",
-- future
x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", 
x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", 
x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", 
x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", 
x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", 
x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", x"0000", x"0000" 
);









--------------------------------------------------------
--      IMPLEMENTATION
--------------------------------------------------------
begin

-- Port A read
process(CLKA)
begin
	if rising_edge(CLKA) then
		if(CSA = '1') then
			DOA_local <= ROM(to_integer(unsigned(ADDRA)));
		else
			DOA_local <= (others => '0');
		end if;
	end if;
end process;

-- tri-state output
DOA <= DOA_local when (CSA = '1') and (OEA = '1') else (others => 'Z');
 

-- Port B read
process(CLKB)
begin
	if rising_edge(CLKB) then
		if(CSB = '1') then
			DOB_local <= ROM(to_integer(unsigned(ADDRB)));
		else
			DOB_local <= (others => '0');
		end if;
	end if;
end process;

-- tri-state output
DOB <= DOB_local when (CSB = '1') and (OEB = '1') else (others => 'Z');

end Behavioral;
